

; memory allocation in heap
; (c) 2017 G.Woigk kio@little-bat.de


; This file implements a memory management with a 'Java'-style garbage collection.
; • Memory is allocated on a heap.
;   • returned memory is always cleared with $00
; • Memory is never actively released.
;   • Implementation of exceptions is therefore very easy. See file 'exceptions.s'
; • Memory is recovered by a Garbage Collection.
;
; ____________________________________________________________________________
; LIMITS of this implementation:
;
; • max. array size: $3fff (flat data), $1ffe (nested data).
; • max. nesting of variables with default setting: 8 levels.
; • Finalizer function not supported.
;
; ____________________________________________________________________________
; Tech NOTES:
;
; block:
;	dw  size + flags
;	db  data[size]
;
; variables point to block.data
;
; flags:
;	bit15	recursion flag during GC, normally 0
;	bit14	1: this block contains nested variables
;	bit13	only if bit14 = 1:
;		0: this is an array
;		1: this is a struct
;		     =>	data[0,1] points to a byte array with offsets: 
;		      	list of offsets (from block.data) of nested variables in this block
;			list must not move during GC (note: could be implemented) => store list in program code
;
; heap:
;	gvar  heap_start: 	heap start address.
;	const heap_end:		heap end address. must be on page boundary, must be ≤ 0xC000.
;	gvar  heap_ptr: 	block allocation pointer. bottom up.
;
; stacks:
;	flat data and return addresses are stored on the machine stack 'rstack'.
;	allocated variables and temporaries are pushed on their own stack 'vstack':
;	gvar vstack_base: 	base address of the vstack. the vstack grows bottom up.
;	reg  BC:		vstack pointer
;
; Requirements:
; • all variables and temporaries, which point to blocks, must be found during GC.
;   • local variables, global variables, array items and struct members.
;   • the 'root' for all variables is the vstack. first item on the vstack is array globals[].
; • heap_end: must be on page boundary, must be ≤ 0xC000, else overflow test must be modified
; • there should always be at least 80 bytes free on the stack to run the GC. 500 bytes free is better!
;   the current code allows a total recursion depth (starting at the locals) of 8 levels.
;   if a program uses longish linked lists, e.g. binary trees, this will cause deeper recursion in the GC. 
;   then the code must be modified and more bytes must be asserted free on the stack.
;
; alloc:
; 	if alloc() can't allocated a block, it will run a garbage collection.
; 	this will take a while.
; 	if alloc still can't allocated a block, it will throw an exception.
; 	during GC, bit15 of block.size may be set.
; 	during GC most blocks will be moved.
; 	during GC variables which point to blocks which are moved are void.
; 	The GC cannot update pointers (or references) to items in struct or arrays.
; 	=> use indexes, to iterate over arrays.
; 	=> assignment to variables in arrays and structs must immediately follow the calculation of their address!
;
; assignment to variables in allocated arrays or structs:
; calculate the address of the variable after the assignment value, and immediately assign the value:
;
;	e.g.:	a[i] = foo();
; 	wrong:	get a; get i; ati; call foo; poke;
;	right:	call foo; get a; get i; ati; set; 
;


#local

#assert lo(heap_end) == 0	; must be on page boundary, must be ≤ 0xC000 (if bottom up)


#code PAGE0
heap_start::	dw  dflt_heap	; base address of heap
heap_ptr::  	dw  dflt_heap	; allocation pointer inside heap
vstack_base:: 	dw  dflt_stacks	; base address of vstack; stack pointer is register BC


#data HEAP_DATA
#assert	$ == HEAP_DATA		; globals.size must be first 
globals_size:	dw  0
globals::			; allocated globals and statics start here



; _________________________________________________
; Initialization code:

#code INIT_GS
	; init globals.size:
	ld	hl, # (heap_data_size - 2) | 0x4000	; bit14=0 => allocated items
	ld	(globals_size), hl

	; setup shadow stack:
	; init vstack_base
	; and push globals[]:
	ld	bc, dflt_stacks + 2
	ld	(vstack_base), bc
	ld	hl, globals
	ld	(dflt_stacks), hl



; _________________________________________________
; data for Garbage Collection

#data DATA

addr_bu	= gc_buffers		; temp. array addr[bsz]
offs_bu	= gc_buffers + $80      ; temp. array offs[bsz]
bsz	= $80                   ; array size

;bu_addr:	dw	0	; temp. array addr[bsz]
;bu_offs:	dw	0	; temp. array offs[bsz]
;busz:		db	0	; array size

addr_ptr:	dw	0	; ptr inside addr[] during block compaction
offs_ptr:	dw	0	; ptr inside offs[] during block compaction
gc_n:		db	0	; remaining size counter



; ––––––––––––––––––––––––––––––––––––––
; helpers:

#include "lib/gc_bsearch.s"



; ––––––––––––––––––––––––––––––––––––––––––
;		Allocate Memory
; ––––––––––––––––––––––––––––––––––––––––––

#code HEAP

; ––––––––––––––––––––––––––––––––––––––
; allocate memory in heap:
;
; alloc ( de:size -- hl:addr de:size )	130cc

alloc::	ld  	hl, (heap_ptr)

	inc 	hl
	add 	hl, de			; hl = block_end - 1

	ld	a,h
	cp	hi(heap_end)		; (block_end-1).hi - heap_end.hi
	jr	nc, 1$			; (block_end-1).hi ≥ heap_end.hi  =>  oomem

	inc 	hl			; hl = block_end
	ld  	(heap_ptr), hl
	dec	hl
	
	sbc	hl,de			; hl = blockend-1-size-1 = block.size

	ld  	(hl++), e
	ld  	(hl++), d
	ret

1$:	push	de
	call	compact
	pop	de
	jr	nz,alloc
	
	ld	a, ERROR_OUT_OF_MEMORY
	jp	throw


; ––––––––––––––––––––––––––––––––––––––
; allocate memory in heap:
;
; alloc ( de:size -- de:addr )		113cc

; NOTE: wenn wir immer 2 bytes frei lassen, kann alloc vereinfacht werden.
; allerdings erhalten wird die block_size nicht mehr.

alloc::	ld  	hl, (heap_ptr)		; de=len, hl->block.size
	ld  	(hl++), e
	ld  	(hl++), d		; de=len, hl->block.data
	
	ex	hl,de			; de -> block.data, hl=size
	add 	hl,de			; de -> block.data, hl -> block_end
	
	inc	hl			; hl = block_end + 1
	ld	a,h
	dec 	hl			; hl -> block_end
	
	sub	a, hi(heap_end)		; (block_end+1).hi - heap_end.hi
	jr	nc, 1$			; (block_end+1).hi ≥ heap_end.hi  =>  oomem

	ld  	(heap_ptr), hl
	ret

1$:	;and	a
	sbc	hl,de			; hl = block_end - block.data = blocksize
	push	hl			; requested size
	xor	a
	ld	(--de),a		; die bereits eingetragene Blocksize wieder löschen
	ld	(--de),a
	call	compact
	pop	de
	jr	nz,alloc

	ld	a, ERROR_OUT_OF_MEMORY
	jp	throw



; -------------------------------------------------
; Garbage Collection
; bottom up heap

compact::
	push	bc		; shadow stack pointer
	
; setup the shadow stack so that it looks like an array of allocated items:
; note: if shadow stack is aligned and limited to page boundary, then only setup of low byte may be required
	ld	hl, bc		; hl -> shadow stack end
	ld	de, (vstack_base)
	dec	hl
	scf
	sbc	hl, de		; hl = shadow stack size, de -> shadow stack block.size
	ex	hl,de		; de = shadow stack size, hl -> shadow stack block.size
	set	6,d		; bit14=1 => allocated items
	ld	(hl++), e
	ld	(hl), d		; set shadow stack block.size
	
; set bit15 in all used blocks
; TODO: evtl. globals nicht als array (mit block.size) ausführen 
; 	und statt dessen die Rekursionsfunktion direkt mit array-addresse und size aufrufen
	ld	ix, gc_set_bit15	; payload function
	ld	hl,bc			; hl -> array.end
	ld	de, (vstack_base)	; de -> array.start
	call	gc_recursion_range_nz	; doit 

; allocate 2 helper arrays:
;	bu_addr[] 	stores addresses of block starts
;	bu_offs[]	stores accumulated gap size == move distance
;	busz		size of each table (size, not count!)

	pop	bc		; shadow stack pointer
	push	bc
	ld	hl, -8*4*2 -4*4	; approx. req. stack size for recursion ((8 calls, 4 words each. **TODO**))
	add	hl, sp
	and	a
	sbc	hl,bc		; hl = guaranteed free space between stacks minus min. table size (huh huh…)
	jp	c, HELP		; negative => sio out buffer, free space in heap, save 512 bytes to disk, …

	ld	de,4*4		; min. table size
	add	hl,de		; hl = free space between stacks
	
	ld	a,h
	cp	2
	jr	c, 5$		; hl < $200 -> won't exceed max. table size (after round down) -> ok
	ld	hl, 254*2	; else 254 bytes per array is maximum (we'll work with byte indexes later)
5$:	srl	h
	rr	l		; hl /= 2	for 2 tables
	res	0,l		; table contains words -> 2byte/entry -> size of each table must be even
	ld	a,l		; a = l = table size
	ld	h,0		; hl = table size	
	add	hl,bc		; hl = shadow stack pointer + table size

	ld	(bu_offs), bc	; one table at shadow stack end
	ld	(bu_addr), hl	; the other behind that table
	ld	(busz),  a	; store table size


; store addresses of used_block_ranges in addr[]
; store accumulated_gap_size before used_block_ranges in offs[]
; move used_block_range down by accumulated_gap_size
; 	either we need a first entry: addr=0,offs=0 for all data before the first gap
; 	or we must test for *var < bu_addr[0] before calling gc_bsearch
; 	we need a last entry: addr=end_of_last_moving_block,offs=0 for all data after heap_ptr
;	so for doing anything at all, tables must have space for at least 2 (or 3) entries. the above code guarantees 4.

	; hl = addr = (bu_addr)
	; de = offs = (bu_offs)
	; a = n = bsz		slots remaining

	; 1st entry: non-moving part before all
	dec	a		; n--
	ld	(gc_n),  a	; store table size		...
	xor	a
	ld	(bc++),a
	ld	(bc++),a	; bu_offs[0] = 0
	ld	(hl++),a
	ld	(hl++),a	; bu_addr[0] = 0
	ld	(addr_ptr), hl				...
	ld	(offs_ptr), bc				...

	; find 1st gap:
	ld	hl,(heap_start)	; hl:q   = (heap_start)
	ld	bc,(heap_ptr)	; bc:end = (heap_ptr)
	jr	2$		; <- this is only required if the heap can be completely empty!
	
	;q += (*q.w++)
4$:	ld	e,(hl++)
	ld	d,(hl++)	; de=block.size, hl->block.data
	bit	7,d
	jr	z,3$		; bit15 not set => unused => break
	res	7,d
	bit	6,d
	jr	z,1$		; flat
	res	6,d		; else clear bit13 and bit14
	res	5,d
1$:	add	hl,de		; hl -> next block
	;jr	2$

	;while hl:q < (bc:heap_ptr) && block(hl:q) is used
2$:	ld	a,h
	cp	b
	jr	nz,4$
	ld	a,l
	cp	c
	jr	nz,4$
	
	; hl == bc => no unused block!
	pop	bc		; shadow stack pointer
	ret			; return z

; --------------
; non-moving used blocks have been skipped
; now skip unused blocks
; hl -> unused block.data
; hl < heap_ptr
; bc = heap_ptr

3$:	;ld	de,hl		; de:z = hl:q	=> gap size = 0
	push	hl;de
	jr	14$

; ---------------
; entry from copy_used_blocks_loop:
; unused block seen => skip over unused blocks
; hl:q -> unused block.data - 1
; de:z -> gap start
; bc = heap_ptr ? NO

23$:	inc	hl		; hl -> unused block.data
	ld	bc, (heap_ptr)
	push	de
	jr	12$
	
	;while block(hl:q) is unused
	;q += (*q.w++)
; hl -> unused block.data
14$:	ld	e,(hl++)	; bc = block.size incl. flags
	ld	d,(hl++)	; hl -> block.data
	bit	7,d
	jr	nz,13$		; bit15 set => used => break
	bit	6,d
	jr	z,11$		; flat
	res	6,d		; else clear bit13 and bit14
	res	5,d
11$:	add	hl,de
	;jr	12$

	;Schleife über unused blocks:
	;while hl:q < (bc:heap_ptr) && 
12$:	ld	a,c
	cp	a,l
	jr	nz,14$
	ld	a,b
	cp	a,h
	jr	nz, 14$

	; break: hl == bc => last block!
	; break by heap_ptr
	pop	de		; de:z -> start of gap
	jr	25$

13$:	; break by used block
	pop	de		; de:z
	
	; test N
	ld	a,(gc_n)
	cp	1
	jr	z, BREAK	; arrays full => must move remainder en block incl. all gaps
	dec	a
	ld	(gc_n),a

	; de = z -> start of gap
	; hl = q -> 1st used block 
	push	de		; z -> start of gap
	push	hl		; q -> 1st used block 
	
	; store offs: (neg.!)	
	ex	hl,de
	and	a
	sbc	hl,de		; d = z - q
	ex	hl,de
	ld	hl,(offs_ptr)
	ld	(hl++),e	; note: offs must be subtracted later
	ld	(hl++),d
	ld	(offs_ptr),hl

	; store addr
	pop	de		; de:q -> 1st used block 
	ld	hl,(addr_ptr)
	ld	(hl++),e
	ld	(hl++),d
	ld	(addr_ptr),hl
	ex	hl,de		; hl:q -> 1st used block
	pop	de		; de:z -> gap
	;jr	24$
	
	; while q < (heap_ptr) && block(q) is used 
	; -> move block data, skip over block
	
24$:	ld	c,(hl++)	; hl -> block.data - 1
	ld	b,(hl)		; bc = block.size incl. flags
	bit	7,b
	jr	z,23$		; bit15=0 => unused => break: goto skip_unused_blocks
	ld	(de++),c
	res	7,b
	bit	6,b
	jr	z,21$		; flat
	res	6,b		; else clear bit13 and bit14
	res	5,b
21$:	inc	bc		; bc must be != 0, therefore ldir also copies the size.hi byte
	ldir	
	;jr	22$

22$:	ld	a,(heap_ptr)
	cp	l
	jr	nz,24$
	ld	a,(heap_ptr+1)
	cp	h
	jr	nz,24$

	; -------------
	; hl:q reached heap_ptr => finished!
	; -------------
	
	; hl:q == bc:end => last block! => break
25$:	ld	(heap_ptr),de	; (heap_ptr) = de:z

	; store final addr, offs and bsz:
	push	hl

	; store addr
	ld	hl,(addr_ptr)
	ld	(hl++),e		
	ld	(hl),d

	; store offs:
	xor	a
	ld	hl,(offs_ptr)
	ld	(hl++),a
	ld	(hl),a

	; update busz:	
	; bsz = bsz - (n+1)
	ld	hl,busz
	ld	a,(gc_n)
	cpl	a		; a = 255 - a = -(a+1)
	add	a,(hl)		; a = bsz - (n+1)
	ld	(hl),a

	pop	hl

	; memset(z, 0, q-z)	
	xor	a
	sbc	hl,de		; hl = q-z = count
	ex	hl,de		; de=cnt, hl=z
	call	memclr		; ( hl:addr de=cnt a:byte -- hl:pres de:0 a:pres )  0-safe!

	; clear bit15 in all used blocks and update addresses in all variables:
	ld	ix, gc_add_offs_and_clr_bit15
	pop	hl
	push	hl			; hl = shadow stack ptr -> array.end
	ld	de, (vstack_base)	; de -> array.start
	
	call	gc_recursion_range_nz	; doit 

	pop	bc		; shadow stack ptr
	ret

















; ––––––––––––––––––––––––––––––––––––––––––––––––––––––––
; variables are visited using a recursion helper function.
; ––––––––––––––––––––––––––––––––––––––––––––––––––––––––

; enter here with:
; hl -> array.end
; de -> array.data
; ix -> payload function

gc_recursion_range_nz::
l1:	ld	b,(--hl)
	ld	c,(--hl)	; hl -> var, bc -> block.data
	ld	a,b
	or	c
	call	nz,JP_IX	; call if not null ( bc->block.data -- hl,de:pres )

gc_recursion_range::
	ld	a,l
	cp	e
	jr	nz,l1		
	ld	a,h
	cp	d
	jr	nz,l1		; while hl != de
	ret

JP_IX:	jp	ix




; ________________________________________________________
; set bit15 in all used variables.
;
; in:	ix -> this function
;	hl -> var
;	bc -> block.data = peekw(hl:var) != null
; out:	-
; pres:	ix, iy, hl, de

gc_set_bit15:
	ld	a,(--bc)	; bc -> block.size.hi, a = block.size.hi
	xor	a,0x80		; toggle bit15
	ret	p		; bit15 was set => already visited => exit
	ld	(bc),a		; set bit15
	
	add	a
	ret	p		; bit14 = 0 => no recursion
	jr	gc_recursion


; ________________________________________________________
; adjust address in variables in range bu_addr[0] … bu_addr[busz] 
; and clear bit15 in all used variables.
;
; in:	ix -> this function
;	hl -> var
;	bc -> block.data = peekw(hl:var) != null
; out:	-
; pres:	ix, iy, hl, de

gc_add_offs_and_clr_bit15:
	; test whether in range:
	; and add offset:	
	push	de		; stack: de
	push	hl		; stack: ->var

	ld	hl,(bu_addr)
	ld	a,(busz)
	;ld	bc,bc		; bc -> block.data
	call	gc_bsearch	; ( hl:array a:size bc:block -- hl:addr[i] )  bc:pres
	res	0,l		; hl.bit0 is void
	ld	e,(gc_n)
	ld	d,0
	add	hl,de
	add	hl,de		; hl -> offs[i]
	ld	e,(hl++)
	ld	d,(hl)		; de = offset (neg.!)
	ex	hl,de		; hl = offset
	add	hl,bc		; hl -> block.data (new)
	ld	bc,hl		; bc -> block.data (new)

	pop	hl		; hl->var
	ld	(hl++),e
	ld	(hl--),d	; store new address, hl: preserved

	pop	de		; de: preserved
	;jr	gc_clr_bit15
	

; clear bit15 in all used variables.
;
; in:	ix -> this function
;	hl -> var
;	bc -> block.data = *var != null
; out:	-
; pres:	ix, iy, hl, de

gc_clr_bit15:
	dec	bc		; bc -> block.size.hi
	ld	a,(bc)
	xor	a,0x80
	ret	m		; bit15 = 0 => already visited
	ld	(bc),a		; set bit15
	
	add	a
	ret	p		; bit14 = 0 => no recursion
	;jr	gc_recursion


; recursion helper:
; the payload function has been executed on this var/block and discovered need for recursion:
; in:	bc -> block.size.hi
;	a  =  block.size.hi << 1
;	ix -> payload function
;	gc_i = the var's index in it's parent struct
; pres:	ix, iy, hl, de

gc_recursion:
	push	hl		; pres.
	push	de		; pres.

	add	a
	jp	p, gc_array	; size.bit13 = 0 => array

; size.bit13 = 1 => struct:
	
	TODO
	
;	offsets = peekw(block+0);
;	n = offsets.size
;	a = offsets
;	e = a + n
;	while(a<e)
;	{
;	    if(a!=i)
;	        var = block+peek(a);
;	        blk = peekw(var);
;	        if(blk!=0) call IX(var, blk, fu, a);
;	    a += 1
;	}
;	if(i<n) 
;	    var = block+peek(i);
;	    block = peekw(var)
;	    if(block!=0) JP IX; 


; size.bit13 = 0 => array
;	stack: hl, bc (pres.)
;	de -> block.size.hi
;	ix -> payload function
gc_array:		
	ex	hl,de		
	dec	hl		; hl -> block.size.lo	
	ld	e,(hl++)	; e = size.lo
	ld	b,e
	ld	a,(hl)	
	inc	hl		; hl -> block.data
	and	0x1F		; a = size.hi	
	ld	d,a		; de = size	
	rra	
	ld	c,a
	rr	b		; cb = array.count

	add	hl,de		; hl -> block.end

	inc	c		; for easier loop end test
	jr	2$
	
	; hl -> behind current item
	; bc = loop counter

1$:	ld	d,(--hl)
	ld	e,(--hl)	; hl -> var, de -> block.data
	ld	a,d
	or	e
	call	nz,JP_IX	; call if not null 

2$:	djnz	1$
	dec	c
	jr	nz,1$
	
	pop	de
	pop	hl
	ret





#endlocal






